/* osgVisual test. distortionNG, experimental.
*
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the "Software"), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*
*  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
*  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
*  THE SOFTWARE.
*/

#include "DistortionManipulator.h"

#include<osgViewer/Viewer>

using namespace osg;
using namespace osgViewer;


DistortionManipulator::DistortionManipulator()
{
	activeSetupMode = DISABLED;
	activeDistortionMode = MESH;
	activeManualSetupMode = DISTORTION;
	activeVisualizationMode = NONE;

	_distortionSet = NULL;
}

DistortionManipulator::~DistortionManipulator()
{
}

void DistortionManipulator::getUsage(osg::ApplicationUsage& usage) const
{
    usage.addKeyboardMouseBinding("Keypad 7","Show distortion mesh / intensity map / none.");
	usage.addKeyboardMouseBinding("Keypad 8","Save distortion set.");	// via plugin
	usage.addKeyboardMouseBinding("Keypad 4","Toggles Setup Mode between DISABLED, MANUAL & DELEGATED.");
 	usage.addKeyboardMouseBinding("Keypad 5","MANUAL Mode: Toggle between blending & distortion setup.");
	usage.addKeyboardMouseBinding("Keypad 6","MANUAL Mode: Toggle if distortion drags affect mesh or rtt texture coordinates.");	// Defaults to Mesh
	usage.addKeyboardMouseBinding("Keypad 1","Reset distortion.");
	usage.addKeyboardMouseBinding("Keypad 2","Reset intensity blending.");
}

bool DistortionManipulator::handle(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter& aa, osg::Object* obj, osg::NodeVisitor* nv)
{
	//osgViewer::View* viewer = dynamic_cast<osgViewer::View*>(&aa);
	//if ( viewer && !_distortionSet.isValid())
	//{
	//	OSG_ALWAYS<<"Fetching valid DistortionSet!"<<std::endl;
	//	_distortionSet = viewer->getDistortionSet();
	//}

	switch(ea.getEventType())
    {
        case(osgGA::GUIEventAdapter::MOVE): break;
        case(osgGA::GUIEventAdapter::DRAG):
		{
			OSG_ALWAYS<<"DRAG!"<<std::endl;
			break;
		}
        case(osgGA::GUIEventAdapter::PUSH):
		{
			if(ea.getButton() == osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON)
			{
				OSG_ALWAYS<<"mouse click left!"<<std::endl;
			}
			break;
		}
        case(osgGA::GUIEventAdapter::RELEASE): break;
        case(osgGA::GUIEventAdapter::KEYDOWN):
		{
			if (ea.getKey()==osgGA::GUIEventAdapter::KEY_KP_End)	// KP 1: reset distortion
			{
				resetDistortion();
				return(true);
			}
			if (ea.getKey()==osgGA::GUIEventAdapter::KEY_KP_Down)	// KP 2: reset intensity map
			{
				resetIntensityMap();
				return(true);
			}
			if (ea.getKey()==osgGA::GUIEventAdapter::KEY_KP_Left)	// KP 4: Toggles Setup Mode between DISABLED, MANUAL & DELEGATED.
			{
				switch(activeSetupMode)
				{
					case DISABLED : activeSetupMode = MANUAL; break;
					case MANUAL : activeSetupMode = DELEGATED; break;
					case DELEGATED : activeSetupMode = DISABLED; break;
				}

				OSG_ALWAYS<<"KEY_KP_4 : Setup Mode now "<<activeSetupMode<<std::endl;
				return(true);
			}
			if (ea.getKey()==osgGA::GUIEventAdapter::KEY_KP_Begin)	// KP 5: MANUAL Mode: Toggle between blending & distortion setup.
			{

				activeManualSetupMode = (activeManualSetupMode==DISTORTION?BLENDING:DISTORTION);
				OSG_ALWAYS<<"KEY_KP_5 : activeManualSetupMode is now "<<activeManualSetupMode<<std::endl;
				return(true);
			}
			if (ea.getKey()==osgGA::GUIEventAdapter::KEY_KP_Right)	// KP 6: MANUAL Mode: Toggle if distortion drags affect mesh or rtt texture coordinates -> defaults to Mesh
			{
				activeDistortionMode = (activeDistortionMode==MESH?TEXCOORDINATES:MESH);
				OSG_ALWAYS<<"KEY_KP_6 : activeDistortionMode is now "<<activeDistortionMode<<std::endl;
				return(true);
			}
			if (ea.getKey()==osgGA::GUIEventAdapter::KEY_KP_Home)	// KP 7: Show distortion mesh / intensity map / none.
			{
				switch(activeVisualizationMode)
				{	
					case DISTORTION_MESH : 
					{
						activeVisualizationMode = INTENSITY_MAP;
						showDistortionMesh(true);
						showIntensityMap(false);
						break;
					}
					case INTENSITY_MAP :
					{
						activeVisualizationMode = NONE;
						showDistortionMesh(false);
						showIntensityMap(true);
						break;
					}
					case NONE : 
					{
						activeVisualizationMode = DISTORTION_MESH;
						showDistortionMesh(false);
						showIntensityMap(false);
						break;
					}
				}


				OSG_ALWAYS<<"KEY_KP_7 : activeVisualizationMode is now "<<activeVisualizationMode<<std::endl<<std::endl;
				return(true);
			}
			if (ea.getKey()==osgGA::GUIEventAdapter::KEY_KP_Up)	// KP 8: Save distortion set	 via plugin
			{
				OSG_ALWAYS<<"KEY_KP_8 : todo: Save DistortionContainer"<<std::endl;
				return(true);
			}
			

			return false;	// Event ignored
		}
        case(osgGA::GUIEventAdapter::KEYUP): break;
		case(osgGA::GUIEventAdapter::FRAME):
		{
			//OSG_ALWAYS<<"FRAME!"<<std::endl;
			break;
		}
        case (osgGA::GUIEventAdapter::RESIZE):	// todo: adapt distortion mesh to new screen size
		{
			OSG_ALWAYS<<"RESIZE!"<<std::endl;
			break;
		}

        default:
            return false;
    }
    return false;
}

void DistortionManipulator::resetIntensityMap()
{
	if(!_distortionSet.valid())
		return;

	osg::ref_ptr<osg::Image> image = _distortionSet->getIntensityMap();

	OSG_ALWAYS<<"Reseting IntensityMap Blending."<<std::endl;

	if (!image->isDataContiguous())
	{
		OSG_WARN<<"Warning: DistortionManipulator does not support working with non contiguous imagery as blendmaps!"<<std::endl;
		return;
	}

	// Fill intensity map with 255 
	unsigned char* dataPtr = image->data(); 
	for(unsigned int i=0;i<image->getTotalSizeInBytes();i++)
	{
		*dataPtr++ = 255; 
	}
	image->dirty();
}

void DistortionManipulator::resetDistortion()
{
	if(!_distortionSet.valid())
		return;

	OSG_ALWAYS<<"ToDo: resetDistortion()"<<std::endl;
}

void DistortionManipulator::showDistortionMesh(bool show)
{
	OSG_ALWAYS<<"ToDo: showDistortionMesh(bool) is now "<<show<<std::endl;

	// Todo: Stateset muss das von der MeshGeode sein.

   /* osg::PolygonMode* polyModeObj = dynamic_cast<osg::PolygonMode*>(_stateset->getAttribute(osg::StateAttribute::POLYGONMODE));
    if (!polyModeObj)
    {
        polyModeObj = new osg::PolygonMode;
        _stateset->setAttribute(polyModeObj);
    }
 
	if(show)
		polyModeObj->setMode(osg::PolygonMode::FRONT_AND_BACK,osg::PolygonMode::LINE);
	else
		polyModeObj->setMode(osg::PolygonMode::FRONT_AND_BACK,osg::PolygonMode::FILL);*/
}

void DistortionManipulator::showIntensityMap(bool show)
{
		OSG_ALWAYS<<"ToDo: showIntensityMap(bool) is now "<<show<<std::endl;

		if(_distortionSet.valid())
			_distortionSet->setShowIntensityMapOnly(show);
}
